#ifdef STREAM_ANALYZER
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <unistd.h>
#include <stdlib.h>
#include "lib.h"
#include "util.h"
#include "streams.h"
#include "streamstat.h"
#include "colors.h"

static inline int is_better_host(int ctype, int counter, int order, sta_host *h, sta_host *than)
{
	if(ctype == COUNTER_PACKET)
	{
		if(order == ORDER_ASC)
			return(h->packets[counter] > than->packets[counter]);
		else
			return(h->packets[counter] < than->packets[counter]);
	}
	else // COUNTER_BYTE
	{
		if(order == ORDER_ASC)
			return(h->bytes[counter] > than->bytes[counter]);
		else
			return(h->bytes[counter] < than->bytes[counter]);
	}
}


static inline void arrayhandle(int *n, void **list, void *current, int maxitems, int c, int ov, int o,
							  int (*better)(int, int, int, void *, void **))
{
	int j;
	
	if(!*n)
	{
		list[(*n)++] = current;
		return;
	}
	if(*n == maxitems)
		if(!better(c, ov, o, current, list[*n - 1])) return;

	for(j = 0; j < *n; j++)
	{
		if(better(c, ov, o, current, list[j]))
		{
			memmove(list + j + 1, list + j, sizeof(void*) * (*n - j - (*n == maxitems ? 1 : 0)));
			list[j] = current;
			break;
		}
	}
	if(j == *n && *n < maxitems)
	{
		int k;
		list[(*n)++] = current;
	}
	return;
}


sta_host ** stastat_gethostlist(unsigned int maxitems, int * count, int ctype, int counter,
								int order, int only_ours)
{
	sta_host ** list = (sta_host**) sta_malloc(sizeof(sta_host*) * maxitems);
	sta_host * current;
	sta_host ** hosthash = stastream_gethosthash();
	int n = 0, i;
	
	
	for(i = 0; i < HOSTHASH_SIZE; i++)
	{
		for(current = hosthash[i]; current; current = current->hh_next)
		{
			int j;
			if(only_ours)
				if(!stastream_isours(current->ip)) continue;
				
			arrayhandle(&n, (void**) list, current, maxitems, ctype, counter, order, 
						(int (*)(int, int, int, void*, void**)) is_better_host);
		}
	}	
	*count = n;
	return(list);
}

static inline int is_better_stream(int ctype, int overall, int order, sta_stream *s, sta_stream *than)
{
	if(ctype == COUNTER_PACKET)
	{
		if(order == ORDER_ASC)
			if(overall)
				return((s->endpoints[0].packets + s->endpoints[1].packets) >
					   (than->endpoints[0].packets + than->endpoints[1].packets));
			else
				return((s->endpoints[0].lastpackets + s->endpoints[1].lastpackets) >
					   (than->endpoints[0].lastpackets + than->endpoints[1].lastpackets));
		else
			if(overall)
				return((s->endpoints[0].packets + s->endpoints[1].packets) <
					   (than->endpoints[0].packets + than->endpoints[1].packets));
			else
				return((s->endpoints[0].lastpackets + s->endpoints[1].lastpackets) <
					   (than->endpoints[0].lastpackets + than->endpoints[1].lastpackets));
	}
	else // COUNTER_BYTE
	{
		if(order == ORDER_ASC)
			if(overall)
				return((s->endpoints[0].bytes + s->endpoints[1].bytes) >
					   (than->endpoints[0].bytes + than->endpoints[1].bytes));
			else
				return((s->endpoints[0].lastbytes + s->endpoints[1].lastbytes) >
					   (than->endpoints[0].lastbytes + than->endpoints[1].lastbytes));
		else
			if(overall)
				return((s->endpoints[0].bytes + s->endpoints[1].bytes) <
					   (than->endpoints[0].bytes + than->endpoints[1].bytes));
			else
				return((s->endpoints[0].lastbytes + s->endpoints[1].lastbytes) <
					   (than->endpoints[0].lastbytes + than->endpoints[1].lastbytes));
	}
}


sta_stream ** stastat_getstreamlist(unsigned int maxitems, int * count, int ctype, int overall,
								int order, int only_ours)
{
	sta_stream ** list = (sta_stream**) sta_malloc(sizeof(sta_stream*) * maxitems);
	sta_stream * current;
	sta_stream ** streamhash;
	int n = 0;
	
	for(current = stastream_getstreamlist(); current; current = current->stal_next)
	{
		int j;
		if(current->properties & SPROP_CLOSED) continue;
		if(only_ours)
			if(!(stastream_isours(current->endpoints[0].host->ip) ||
				 stastream_isours(current->endpoints[1].host->ip))) continue;

		arrayhandle(&n, (void**) list, current, maxitems, ctype, overall, order,
					(int (*)(int, int, int, void*, void**)) is_better_stream);
	}

	*count = n;
	return(list);
}

static inline void stastat_calcproto(sta_stream * head, struct protoinfo * pi, int id)
{
	memset(pi, 0, sizeof(struct protoinfo));
	pi->protoid = id;
	for(; head; head = head->proto_next)
	{
		pi->counters[PC_OVERALL_BYTES] += head->endpoints[0].bytes + head->endpoints[1].bytes;
		pi->counters[PC_OVERALL_PACKETS] += head->endpoints[0].packets + head->endpoints[1].packets;
		pi->counters[PC_LAST_BYTES] += head->endpoints[0].lastbytes + head->endpoints[1].lastbytes;
		pi->counters[PC_LAST_PACKETS] += head->endpoints[0].lastpackets + head->endpoints[1].lastpackets;
	}
}

static inline int is_better_protocol(int counter, int pad, int order, struct protoinfo *pi, struct protoinfo *than)
{
	if(order == ORDER_ASC)
		return(pi->counters[counter] > than->counters[counter]);
	else
		return(pi->counters[counter] < than->counters[counter]);
}


struct protoinfo ** stastat_getprotolist(unsigned int maxitems, int * count, int ctype, int overall,
								int order)
{
	int i, counter, n = 0;
	static struct protoinfo pi[256];
	struct protoinfo ** list = (struct protoinfo**) sta_malloc(sizeof(struct protoinfo*) * maxitems);

	for(i = 0; i < STREAM_MAX; i++)
		stastat_calcproto(stastream_getprotolist(i), &pi[i], i);
	
	if(ctype == COUNTER_PACKET)
		if(overall)
			counter = PC_OVERALL_PACKETS;
		else
			counter = PC_LAST_PACKETS;
	else
		if(overall)
			counter = PC_OVERALL_BYTES;
		else
			counter = PC_LAST_BYTES;
	
	for(i = 0; i < STREAM_MAX; i++)
		arrayhandle(&n, (void**) list, &pi[i], maxitems, counter, 0, order, 
						(int (*)(int, int, int, void*, void**)) is_better_protocol);

	*count = n;

	return(list);
}


int stastat_compare(const void *a, const void *b)
{
	sta_stream *as = *((sta_stream**) a);
	sta_stream *bs = *((sta_stream**) b);
	int p1, p2, q1, q2;
	
	/*  1 - a ma by w tablicy za b */
	/* -1 - b ma by w tablicy za a */

	if((as->properties & SPROP_STALE) && (!(bs->properties & SPROP_STALE))) return( 1);
	if((bs->properties & SPROP_STALE) && (!(as->properties & SPROP_STALE))) return(-1);
	
	p1 = as->properties & SPROP_PROTO_MASK;
	p2 = bs->properties & SPROP_PROTO_MASK;
	
	if(p1 == SPROP_PROTO_TCP && p2 != SPROP_PROTO_TCP) return(-1);
	if(p2 == SPROP_PROTO_TCP && p1 != SPROP_PROTO_TCP) return( 1);
	if(p1 == SPROP_PROTO_UDP && p2 != SPROP_PROTO_UDP) return(-1);
	if(p2 == SPROP_PROTO_UDP && p1 != SPROP_PROTO_UDP) return( 1);
	
	q1 = as->endpoints[0].bytes + as->endpoints[1].bytes;
	q2 = bs->endpoints[0].bytes + bs->endpoints[1].bytes;
	
	if(q1 > q2)
		return(-1);
	else if(q1 < q2)
		return(1);
	else
		return(0);
}

void stastat_sortstreams(sta_stream ** list, int n)
{
	int i, z;
	sta_stream * temp;
	/*
	   TCP streams (by overall bytes)
	   UDP streams (by overall bytes)
	   ICMP streams (by overall bytes)
	   other streams (by overall bytes)
	   stale TCP streams 
	   stale UDP streams
	   stale ICMP streams
	   other streams */
	qsort(list, n, sizeof(sta_stream*), stastat_compare);
}


char * stastat_stream2a(char *indent, sta_stream *s, int verbosity, int human, int color, int tab, double ms)
{
	char endpoints[100];
	char epa[20];
	char epb[20];
	
	char overall_bps[20];
	char overall_pps[20];
	char overall_bytes[20];
	char overall_packets[20];

	struct in_addr in;
	int proto;
	
	double lb, lp;
	char separator;
	
	if(tab)
		separator = '\t';
	else
		separator = ' ';
	
	in.s_addr = s->endpoints[0].host->ip;
	snprintf(epa, 20, "%s", inet_ntoa(in));
	in.s_addr = s->endpoints[1].host->ip;
	snprintf(epb, 20, "%s", inet_ntoa(in));

	proto = s->properties & SPROP_PROTO_MASK;
	if(proto == SPROP_PROTO_TCP || proto == SPROP_PROTO_UDP)
	{
		snprintf(endpoints, 100, color ? ((s->properties & SPROP_STALE) ?
										(SCOLOR_PROTO_TCPUDP "%c%c" SCOLOR_HOST
										"%s\033[0m" SCOLOR_PORT ":%d%c"
										SCOLOR_HOST "%s\033[0m" SCOLOR_PORT ":%d\033[0m") :
										(COLOR_PROTO_TCPUDP "%c%c" COLOR_HOST
										"%s\033[0m" COLOR_PORT ":%d%c"
										COLOR_HOST "%s\033[0m" COLOR_PORT ":%d\033[0m")) :
										"%c%c%s:%d%c%s:%d",
										(proto == SPROP_PROTO_TCP) ?
											((s->properties & SPROP_UNIDIRECTIONAL) ? 't' : 'T') :
											((s->properties & SPROP_UNIDIRECTIONAL) ? 'u' : 'U'),
										(s->properties & SPROP_STALE) ? '~' : ' ',										epa, ntohs(s->endpoints[0].port),
										separator,
										epb, ntohs(s->endpoints[1].port));
	}
	else
	{
		snprintf(endpoints, 100, color ? ((s->properties & SPROP_STALE) ?
										(SCOLOR_PROTO_ICMP "%c%c" SCOLOR_HOST "%s%c%s\033[0m") :
										(COLOR_PROTO_ICMP "%c%c" COLOR_HOST	"%s%c%s\033[0m")) :
										"%c%c%s%c%s",
										(proto == SPROP_PROTO_ICMP) ?
											((s->properties & SPROP_UNIDIRECTIONAL) ? 'i' : 'I') :
											((s->properties & SPROP_UNIDIRECTIONAL) ? '.' : '?'),
										(s->properties & SPROP_STALE) ? '~' : ' ',
 										epa,
										separator,
										epb);
	}
	
	lb = s->endpoints[0].lastbytes + s->endpoints[1].lastbytes;
	lp = s->endpoints[0].lastpackets + s->endpoints[1].lastpackets;

	mkhuman(s->endpoints[0].bytes + s->endpoints[1].bytes, 20, overall_bytes, "B", 1024, human);
	mkhuman(s->endpoints[0].packets + s->endpoints[1].packets, 20, overall_packets, "p", 1000, human);
	mkhuman(lb * 1000.0 / ms, 20, overall_bps, "Bps", 1024, human);
	mkhuman(lp * 1000.0 / ms, 20, overall_pps, "pps", 1000, human);

	switch(verbosity)
	{
		case 0:
			return(auto_snprintf(300, color ? ((s->properties & SPROP_STALE) ? 
												("%s%c\033[0m" SCOLOR_DATA "%s/%s%c\033[0m" SCOLOR_RATE "%s/%s\033[0m") :
												("%s%c\033[0m" COLOR_DATA "%s/%s%c\033[0m" COLOR_RATE "%s/%s\033[0m")) :
												"%s%c%s/%s%c%s/%s",
													endpoints, 
													separator,
													overall_bytes, overall_packets, separator,
													overall_bps, overall_pps));
		case 1:		
			return(auto_snprintf(300, color ? ((s->properties & SPROP_STALE) ?
												("%s%c\033[0m" SCOLOR_DATA "%s/%s%c\033[0m" SCOLOR_RATE "%s/%s\033[0m\n%s  " SCOLOR_PROTOCOL "%s\033[0m") :
												("%s%c\033[0m" COLOR_DATA "%s/%s%c\033[0m" COLOR_RATE "%s/%s\033[0m\n%s  " COLOR_PROTOCOL "%s\033[0m")) :
												"%s%c%s/%s%c%s/%s\n%s  %s",
													endpoints, separator,
													overall_bytes, overall_packets, separator,
													overall_bps, overall_pps, indent, sta_protoinfo(s)));
	}
}

sta_stream ** stastat_getstreams(sta_host *h, int *count)
{
	sta_stream ** list;
	sta_stream *c;
	int n, l = h->streamcount;
	
	list = (sta_stream**) sta_malloc(sizeof(sta_stream*) * h->streamcount);
	for(n = 0, c = h->streamlist; n < l && c; )
	{
		list[n++] = c;
		if(h == c->endpoints[0].host)
			c = c->endpoints[0].sl_next;
		else
			c = c->endpoints[1].sl_next;
	}
	*count = n;
	return(list);
}
#endif
